Imports System.Drawing
Imports System.Drawing.Drawing2D
Imports System.Drawing.Imaging
Imports System.Math

Imports UvARescue.Math


Public Class AgentsLayer
    Inherits BufferedLayer

    Private ReadOnly _RAD2DEG As Double = 180.0 / PI
    Private ReadOnly _DEG2RAD As Double = PI / 180.0

#Region " Constructor / Destructor "

    Public Sub New(ByVal parent As ManifoldImage)
        MyBase.New(parent)

        Me._AgentsColor = Color.Black
        Me.ResetAgentsPen()

        Me._PoseColor = Color.Chartreuse '#7FFF00 The path is per definition cleared
        Me.ResetPoseBrush()

    End Sub

    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If disposing Then

            Try

                If Not IsNothing(Me._PoseBrush) Then
                    Me._PoseBrush.Dispose()
                    Me._PoseBrush = Nothing
                End If
                If Not IsNothing(Me._AgentsPen) Then
                    Me._AgentsPen.Dispose()
                    Me._AgentsPen = Nothing
                End If

            Catch ex As Exception
                Console.Error.WriteLine(ex.ToString)

            End Try

        End If

        MyBase.Dispose(disposing)

    End Sub

#End Region

#Region " Brushes "

    Private _PoseColor As Color
    Private _PoseBrush As Brush
    Private _PoseBrush2 As Brush = New SolidBrush(Color.Red)

    Public Property PoseColor() As Color
        Get
            Return Me._PoseColor
        End Get
        Set(ByVal value As Color)
            If Not value = Me._PoseColor Then
                Me._PoseColor = value
                Me.ResetPoseBrush()
            End If
        End Set
    End Property
    Protected ReadOnly Property PoseBrush() As Brush
        Get
            Return Me._PoseBrush
        End Get
    End Property
    Private Sub ResetPoseBrush()
        If Not IsNothing(Me._PoseBrush) Then
            Me._PoseBrush.Dispose()
        End If
        Me._PoseBrush = New SolidBrush(Me._PoseColor)
    End Sub

#End Region

#Region " Maintain Information "

    Private _AgentsColor As Color
    Private _AgentsPen As Pen
    Public Property AgentsColor() As Color
        Get
            Return Me._AgentsColor
        End Get
        Set(ByVal value As Color)
            If Not value = Me._AgentsColor Then
                Me._AgentsColor = value
                Me.ResetAgentsPen()
            End If
        End Set
    End Property
    Protected ReadOnly Property AgentsPen() As Pen
        Get
            Return Me._AgentsPen
        End Get
    End Property
    Private Sub ResetAgentsPen()
        If Not IsNothing(Me._AgentsPen) Then
            Me._AgentsPen.Dispose()
        End If
        Me._AgentsPen = New Pen(Me._AgentsColor)
    End Sub


    Private poses As New Dictionary(Of Guid, Pose2D)

    Private sizes As New Dictionary(Of Guid, Size)
    Private names As New Dictionary(Of Guid, String)

    Public Overrides Sub RenderAgent(ByVal agent As Agent, ByVal pose As Math.Pose2D)
        MyBase.RenderAgent(agent, pose)

        Console.WriteLine(String.Format("[AgentsLayer] PoseEstimate Update {0} @ {1}.{2} sec,", pose.ToString, DateTime.Now().Second, DateTime.Now().Millisecond))

        'EMO

        'Dim i As Guid

        'Dim pt As Vector2
        '    If Not IsNothing(agent.PathPlanGoalPos2) Then

        '    For Each pt In agent.PathPlanGoalPos2
        '        '(ByVal manifold As Manifold, ByVal scan As ScanObservation, ByVal rawOrigin As Pose2D, ByVal agentName As String)
        '        RenderPatch(New Patch(agent.Manifold, New ScanObservation(0.0F, Nothing), New Pose2D(pt(0), pt(1), 0.0F), "test"))
        '        'Console.WriteLine("RenderPatch")
        '    Next pt
        '    End If


        If Not Me.poses.ContainsKey(agent.UniqueID) Then
            Me.poses.Add(agent.UniqueID, pose)
            Me.sizes.Add(agent.UniqueID, agent.Size)
            Me.names.Add(agent.UniqueID, agent.Name)
        Else
            'no need to update sizes and names, only the pose is dynamic
            Me.poses(agent.UniqueID) = pose
        End If

    End Sub

    Friend Overrides Sub NotifyImageReset()
        MyBase.NotifyImageReset()

        Me.poses.Clear()
        Me.sizes.Clear()
        Me.names.Clear()

    End Sub

#End Region

#Region " Rendering "

    Public Overrides Sub RenderPatch(ByVal patch As Patch)
        MyBase.RenderPatch(patch)

        'paint patch origins
        'If patch.AgentName = "test" Then
        '    Me.Gfx.FillEllipse(Me._PoseBrush2, Me.ComputeScaledRect(patch.EstimatedOrigin.Position, 1))
        'Else
            Me.Gfx.FillEllipse(Me._PoseBrush, Me.ComputeScaledRect(patch.EstimatedOrigin.Position, 3))
        'End If
    End Sub

    Public Overrides Sub Draw(ByVal g As Graphics, ByVal renderingMode As ManifoldRenderingMode)
        MyBase.Draw(g, renderingMode)

        If Not IsNothing(Me.poses) AndAlso Me.poses.Count > 0 Then

            'construct graphics path with current agent poses
            Using path As New GraphicsPath

                Dim pose As Pose2D
                Dim size As Size

                For Each key As Guid In Me.poses.Keys
                    pose = Me.poses(key)
                    size = Me.sizes(key)

                    'construct ellipse centered on pose
                    path.StartFigure()
                    path.AddEllipse(New Rectangle(New Point(CInt(pose.X - size.Width / 2), CInt(pose.Y - size.Height / 2)), size))
                    path.CloseFigure()

                    'construct line indicating the rotation
                    Dim len As Single = 2 * Max(size.Height, size.Width)
                    Dim dx As Single = CType(Cos(pose.Rotation) * len, Single)
                    Dim dy As Single = CType(Sin(pose.Rotation) * len, Single)

                    path.StartFigure()
                    path.AddLine(CType(pose.X, Single), CType(pose.Y, Single), CType(pose.X + dx, Single), CType(pose.Y + dy, Single))
                    path.CloseFigure()

                    'also paint the agent's name just outside the circle's radius
                    Me.DrawReadableString(g, renderingMode, New PointF(CType(pose.X + 0.5F * Max(size.Height, size.Width), Single), CType(pose.Y + 0.5F * Max(size.Height, size.Width), Single)), Me.names(key))

                Next

                'transform from woorld coords to page coords
                path.Transform(Me.Parent.TransformationMatrix)

                g.DrawPath(Me._AgentsPen, path)
#If NEEDED Then
                path.Dispose()

                'construct ellipse centered on pose
                path.StartFigure()
                path.AddArc(New Rectangle(New Point(CInt(pose.X - size.Width / 2), CInt(pose.Y - size.Height / 2)), size), CType(pose.Rotation / Me._RAD2DEG, Single), 90)
                path.CloseFigure()

                Dim brush As New SolidBrush(Color.DarkBlue)

                ' The fill mode is FillMode.Alternate by default.
                g.FillPath(brush, path)
                g.DrawPath(Me._AgentsPen, path)
#End If
            End Using

        End If

    End Sub

#End Region

End Class
