//This is a reimplementation of CameraData.vb (rev. 2257)

using System.Drawing;
//using System.Drawing.Drawing2D;
using System.Drawing.Imaging;

using System.Runtime.InteropServices;

using System.IO;
using MathNet.Numerics; //Defines Complex
using MathNet.SignalProcessing.DataSources;


namespace UvArescue.Agent.Sensor 
{


public enum CameraImageType : byte
{
    Raw,
    Jpeg,
    Png // altURI
};

    [Serializable]
public class CameraData : ISensorData // implements ISensorData
{
    //Public Declare Function htonl Lib "wsock32.dll" (ByVal hostshort As Long) As Long
    //Public Declare Function ntohl Lib "wsock32.dll" (ByVal netshort As Long) As Long

    private bool PostponeConvertion = true;

#region " Constructor "

    Public  CameraData()
{
    // nothing
}


#endregion

#region " Properties "

    private CameraImageType _CameraImageType;

    public readonly byte CameraImageType {
        get { return _CameraImageType; }
    }

    private int _Width = 0;
    public readonly int  Width {
        get { return _Width; }
    }

    private int _Height  = 0;
    public readonly int Height {
        get { return _Height; }
    }

    private byte[] _RawData = null;
    public readonly byte[] RawData {
        get {
            if (Me._Bitmap != null && PostponeConvertion = true) {
                Me.ConvertFromBitmap();
            }
            return _RawData;
        }
    }

    private Bitmap _Bitmap  = null;
    public Bitmap Bitmap {
        get { return Me._Bitmap;  }
        set { Me._Bitmap = value; }
    }

    private int _Sequence = 0;
    public readonly int Sequence {
        get { return _Sequence; }
    }


#endregion

#region " Load "

    protected Object Mutex;


    public void Load(byte[] bytes, int viewNumber, byte multiviewMode , int sequenceNumber)
    {

        if (bytes.Length <= 5) {
            return;
        }
        
        Me._Sequence = sequenceNumber;

        if (bytes(0) == 0) {
            Me._CameraImageType = UvARescue.Agent.CameraImageType.Raw;
        } else if (bytes(0) == 6) {
            Me._CameraImageType = UvARescue.Agent.CameraImageType.Png;
        } else {
            Me._CameraImageType = UvARescue.Agent.CameraImageType.Jpeg;
        }

        // convert 4 bytes to Integer
        int length  = CInt(bytes(1) * 2 ^ 24) + CInt(bytes(2) * 2 ^ 16) + CInt(bytes(3) * 2 ^ 8) + CInt(bytes(4) * 2 ^ 0);

        if ( Me._CameraImageType == UvARescue.Agent.CameraImageType.Raw)
        {
            lock (Me.Mutex)
            {

                Me._Width = CInt(bytes(5) * 2 ^ 8) + CInt(bytes(6));
                Me._Height = CInt(bytes(7) * 2 ^ 8) + CInt(bytes(8));

                    //copy raw data RGB(1 byte + 1 byte + 1 byte )
                Me._RawData = New(length - 5);
                byte[] raw = New(length - 5); //A clone which can be discarded

                for (int i = 0; i < Me._RawData.Length - 1; ++i)
                {
                    Me._RawData(i) = bytes(i + 9);
                    raw(i) = bytes(i + 9);
                }

                if( Me._Bitmap != null) {
                    Me._Bitmap.Dispose(); //clean up, to prevent memory problems
                    Me._Bitmap = null;
                }

                if (multiviewMode < 2 && true ) { //SwitchOff for UsarSkinAgent

                    Bitmap bmp = null;

                    if (Me._Bitmap != null )  {
                        bmp = New(Me.Width, Me.Height, PixelFormat.Format24bppRgb);
                    } else {
                        bmp = Me._Bitmap;
                    }

                    // Lock the bitmap's bits.  
                    Rectangle rect = New(0, 0, bmp.Width, bmp.Height);
                    System.Drawing.Imaging.BitmapData bmpData = bmp.LockBits(rect, Drawing.Imaging.ImageLockMode.ReadWrite, bmp.PixelFormat);

                    //Get the address of the first line.
                    IntPtr ptr = bmpData.Scan0;

                    // Declare an array to hold the bytes of the bitmap.
                    // This code is specific to a bitmap with 24 bits per pixels.

                    // Copy the RGB values back to the bitmap
                    System.Runtime.InteropServices.Marshal.Copy(raw, 0, ptr, raw.Length);

                    // Unlock the bits.
                    bmp.UnlockBits(bmpData);


                    Me._Bitmap = DirectCast(bmp.Clone, Bitmap);
                    bmp.Dispose();

                 } else {
                        //get Multiview subview 
                        //Me._RawData = GetSubView(Me._RawData, viewNumber) Not implemented yet
                 }

            } // end lock
        }
        else if ( Me._CameraImageType ==  UvARescue.Agent.CameraImageType.Png)
        {

            lock (Me.Mutex)
            {

                    if (Me._Bitmap != null) {
                        Me._Bitmap.Dispose(); //clean up, to prevent memory problems
                    }

                    //Dim png(CInt(1.33 * length)) As Byte
                    //For i As Integer = 0 To bytes.Length - 9 Step 4
                    //    png(i) = bytes(i + 5 + 1)
                    //    png(i + 1) = bytes(i + 5 + 2)
                    //    png(i + 2) = bytes(i + 5)
                    //    png(i + 3) = 0
                    //Next

                    //Dim png(length - 1) As Byte
                    //For i As Integer = 0 To bytes.Length - 8 Step 3
                    //    png(i) = bytes(i + 5)
                    //    png(i + 1) = bytes(i + 5 + 2)
                    //    png(i + 2) = bytes(i + 5 + 1) 'This works OK, only upside down, wrong colours and small region at the right to the left.
                    //Next

                    Byte png = New(length - 1);
                    for (int i= 0; bytes.Length - 8; i+=3) {
                        png(i) = bytes(bytes.Length - 1 - i - 2);  //magneta (0,2,1) 'bluered(0,1,2), 'bluepink(1,0,2), MagnetaPink (2,0,1), GreenPink (1,2,0), 
                        png(i + 1) = bytes(bytes.Length - 1 - i - 1);
                        png(i + 2) = bytes(bytes.Length - 1 - i);
                    }

                    if (Me._Bitmap != null) {
                        Me._Bitmap.Dispose(); //clean up, to prevent memory problems
                        Me._Bitmap = null;
                    }

                    if (multiviewMode < 2 && true ) { //SwitchOff for UsarSkinAgent

                        Bitmap bmp = null;

                        if (Me._Bitmap == null) {

                            bmp = New(176, 143, PixelFormat.Format24bppRgb); //Is encoded in png-format, maybe also in bmp
                        } else {
                            bmp = Me._Bitmap;
                        }

                        // Lock the bitmap's bits.  
                        Rectangle rect = New(0, 0, bmp.Width, bmp.Height);
                        System.Drawing.Imaging.BitmapData bmpData  = bmp.LockBits(rect, Drawing.Imaging.ImageLockMode.ReadWrite, bmp.PixelFormat);

                        // Get the address of the first line.
                        IntPtr ptr = bmpData.Scan0;

                        // Declare an array to hold the bytes of the bitmap.
                        // This code is specific to a bitmap with 24 bits per pixels.

                        // Copy the RGB values back to the bitmap
                        System.Runtime.InteropServices.Marshal.Copy(png, 0, ptr, png.Length);

                        // Unlock the bits.
                        bmp.UnlockBits(bmpData);


                        Me._Bitmap = DirectCast(bmp.Clone, Bitmap);
                        bmp.Dispose();
                        Me._Bitmap.RotateFlip(System.Drawing.RotateFlipType.RotateNoneFlipX);

                        Me._Width = Me._Bitmap.Width;
                        Me._Height = Me._Bitmap.Height;
                    } else {
                        //get Multiview subview 
                        //Me._RawData = GetSubView(Me._RawData, viewNumber) Not implemented yet
                    }

                } // end lock
        } else {
                //jpeg
                lock (Me.Mutex)
                {
                    if (Me._Bitmap != null) {
                        Me._Bitmap.Dispose(); //clean up, to prevent memory problems
                    }

                    Byte jpeg = New(length - 1);
                    for (int i = 0 ; i < jpeg.Length - 1; i++) {
                        jpeg(i) = bytes(i + 5);
                    }

                    //if Multiview is not in use
                    if ( multiviewMode < 2 ) {
                        //Me._Bitmap = DirectCast(New Bitmap(New System.IO.MemoryStream(jpeg)).Clone, Bitmap);

                    } else {
                        //get Multiview subview 

                        //Late binding, suggested by http://social.msdn.microsoft.com/Forums/en-US/vbgeneral/thread/e849524d-6c79-47b1-8724-651c8ea7dd93
                        System.IO.MemoryStream stream = New(jpeg);
                        Bitmap subView = New(stream, False);


                        //Dim subView As Bitmap
                        //subView = DirectCast(New Bitmap(New System.IO.MemoryStream(jpeg)).Clone, Bitmap)

                        Me._Bitmap = GetSubView(subView, viewNumber, multiviewMode);


                    }

                    Me._Width = Me._Bitmap.Width;
                    Me._Height = Me._Bitmap.Height;

                } // end lock

        }

        }

    //subview getter function for Raw data
    private Byte[] GetSubView(Byte[] rawData, int spawnNumber)
    {

        //TODO: Raw subview getter function should be implemented here


        return null;
    }


    //subview getter function for Jpeg data
    private Bitmap GetSubView(Bitmap bMap, int spawnNumber, byte multiviewMode) 
{

        try
        {
            //hardcopy, suggested by http://bytes.com/topic/visual-basic-net/answers/371925-new-bitmap-very-slow
            //Dim fullView As Bitmap = New Bitmap(bMap.Width, bMap.Height)
            //Dim grx As Graphics = Graphics.FromImage(bMap)
            //grx.DrawImage(fullView, 0, 0)
            //grx.Dispose()

            Bitmap fullView = New(bMap);

            //saving cameradata to file
            //Dim pathName, imageFile As String
            //pathName = Path.Combine(My.Application.Info.DirectoryPath, "Images")
            //imageFile = Path.Combine(pathName, String.Concat(CStr(spawnNumber), "_FULL.jpg"))
            //Dim fileInfo As New FileInfo(imageFile)
            //If Not fileInfo.Directory.Exists Then
            //   FileInfo.Directory.Create()
            //End If
            //fullView.Save(imageFile, System.Drawing.Imaging.ImageFormat.Jpeg)

            int row, column, x, y, width, height;

            //row = (spawnNumber - 1) Mod multiviewMode
            //column = (spawnNumber - 1) \ multiviewMode

            row = mod(spawnNumber, multiviewMode);
            column = (spawnNumber) / multiviewMode;

            if (row < 0) {
                Console.WriteLine("[GetSubView]: Warning row < 0, set to 0");
                row = 0;
            }
            if (column < 0) {
                Console.WriteLine("[GetSubView]: Warning column < 0, set to 0");
                column = 0;
            }

            width = fullView.Width / multiviewMode;
            height = fullView.Height / multiviewMode;

            //Console.WriteLine("full: {0},{1}, cropped: {2},{3}", fullView.Width, fullView.Height, width, height)

            x = row * width;
            y = column * height;

            Drawing.Rectangle cloneRect = New (x, y, width, height);
            Imaging.PixelFormat format  = fullView.PixelFormat;
            Bitmap subView = fullView.Clone(cloneRect, format);

            //saving cameradata to file
            //imageFile = Path.Combine(pathName, String.Concat(CStr(spawnNumber), "_SUBVIEW.jpg"))
            //subView.Save(imageFile, System.Drawing.Imaging.ImageFormat.Jpeg)

            return subView;
        }
        catch (Exception ex)
        {
            Console.WriteLine("Error occurred while trying to create a subview.");
            Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace);
            Return bMap;

        }

    }

    public void ISensorData.Load(System.Collections.Specialized.StringDictionary msg )
    {
    }

    private bool SlowConvertMethod = false;

    private void ConvertFromBitmap()
    {

        //copy data from Me._Bitmap to Me._RawData

         Me._RawData = New(Me._Width * Me._Height * 3);

        if (SlowConvertMethod == true && Me._Bitmap.PixelFormat == PixelFormat.Format24bppRgb ) {

            BitmapData bits = Me._Bitmap.LockBits(New(0, 0, Me._Bitmap.Width, Me._Bitmap.Height), ImageLockMode.ReadOnly, Me._Bitmap.PixelFormat);

            //From example http://www.bobpowell.net/mapfill.aspx

            int offset = 0;
            int index = 0;
            for ( int x  = 0 ; x < Me.Width - 1; x++) {
                for (int y = 0 ; y < Me.Height - 1; y++) {

                    offset = y * bits.Stride + (3 * x); //3 colors x 8 bits = 24bbp
                    _RawData(index) = Marshal.ReadByte(bits.Scan0, offset); //red
                    _RawData(index + 1) = Marshal.ReadByte(bits.Scan0, offset + 1); //green
                    _RawData(index + 2) = Marshal.ReadByte(bits.Scan0, offset + 2); //blue

                    index += 3;
                } //y
            } //x

            Me._Bitmap.UnlockBits(bits);

        } else if (SlowConvertMethod == true && Me._Bitmap.PixelFormat == PixelFormat.Format32bppArgb ) {

            BitmapData bits = Me._Bitmap.LockBits(New(0, 0, Me._Bitmap.Width, Me._Bitmap.Height), ImageLockMode.ReadOnly, Me._Bitmap.PixelFormat);

            //From example http://www.bobpowell.net/mapfill.aspx

           int offset = 0;
            int index = 0;
            for ( int x  = 0 ; x < Me.Width - 1; x++) {
                for (int y = 0 ; y < Me.Height - 1; y++) {

                    offset = y * bits.Stride + (4 * x); //skip alpha

                    _RawData(index) = Marshal.ReadByte(bits.Scan0, offset); //red
                    _RawData(index + 1) = Marshal.ReadByte(bits.Scan0, offset + 1); //green
                    _RawData(index + 2) = Marshal.ReadByte(bits.Scan0, offset + 2); //blue

                    index += 3;
                } //y
            } //x

            Me._Bitmap.UnlockBits(bits);

        } else {
            //Fast, but memory problems for larger resolutions

            Complex R = New(Me._Width * Me._Height); 
            Complex G = New(Me._Width * Me._Height);
            Complex B = New(Me._Width * Me._Height);

            try {
                MathNet.SignalProcessing.DataSources.BitmapConverter.ReadChannel(Me._Bitmap, R, G, B);
            }
            catch ( Exception ex )
            {

                Console.WriteLine("Error occurred while trying to read RGB channels.");
                Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace);
            }

            // ReDim Me._RawData(R.Length * 3) ' already done above

            int index = 0;

            try
            {
                for (int i = 0 ; i < (R.Length) - 1; i++) {
                    _RawData(index) = CByte(R(i).Real * 256);
                    _RawData(index + 1) = CByte(G(i).Real * 256);
                    _RawData(index + 2) = CByte(B(i).Real * 256);
                    index += 3;
                }
            }
            catch ( Exception ex )
            {
                Console.WriteLine("Error converting RGB data to RAW image");
                Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace);
            }


        }
    }

#endregion

} // End Class

} // end namespace
