Imports System.Drawing
'Imports System.Drawing.Drawing2D
Imports System.Drawing.Imaging

Imports System.Runtime.InteropServices

Imports System.IO
Imports MathNet.Numerics 'Defines Complex
Imports MathNet.SignalProcessing.DataSources




Public Enum CameraImageType As Byte
    Raw
    Jpeg
    Png 'altURI
End Enum

Public Class CameraData
    Implements ISensorData

    'Public Declare Function htonl Lib "wsock32.dll" (ByVal hostshort As Long) As Long
    'Public Declare Function ntohl Lib "wsock32.dll" (ByVal netshort As Long) As Long

    Private PostponeConvertion As Boolean = True

#Region " Constructor "

    Public Sub New()
    End Sub

#End Region

#Region " Properties "

    Private _CameraImageType As CameraImageType
    Public ReadOnly Property CameraImageType() As Byte
        Get
            Return _CameraImageType
        End Get
    End Property

    Private _Width As Integer = 0
    Public ReadOnly Property Width() As Integer
        Get
            Return _Width
        End Get
    End Property

    Private _Height As Integer = 0
    Public ReadOnly Property Height() As Integer
        Get
            Return _Height
        End Get
    End Property

    Private _RawData As Byte() = Nothing
    Public ReadOnly Property RawData() As Byte()
        Get
            If Not IsNothing(Me._Bitmap) AndAlso PostponeConvertion = True Then
                Me.ConvertFromBitmap()
            End If
            Return _RawData
        End Get
    End Property

    Private _Bitmap As Bitmap = Nothing
    Public Property Bitmap() As Bitmap
        Get
            Return Me._Bitmap
        End Get
        Set(ByVal value As Bitmap)
            Me._Bitmap = value
        End Set
    End Property

    Private _Sequence As Integer = 0
    Public ReadOnly Property Sequence() As Integer
        Get
            Return _Sequence
        End Get
    End Property


#End Region

#Region " Load "

    Protected Mutex As New Object


    Public Sub Load(ByVal bytes As Byte(), ByVal viewNumber As Integer, ByVal multiviewMode As Byte, ByVal sequenceNumber As Integer)

        If bytes.Length <= 5 Then
            bytes = Nothing
            Return
        End If

        Me._Sequence = sequenceNumber

        If bytes(0) = 0 Then
            Me._CameraImageType = UvARescue.Agent.CameraImageType.Raw
        ElseIf bytes(0) = 6 Then
            Me._CameraImageType = UvARescue.Agent.CameraImageType.Png
        Else
            Me._CameraImageType = UvARescue.Agent.CameraImageType.Jpeg
        End If

        ' convert 4 bytes to Integer
        Dim length As Integer = CInt(bytes(1) * 2 ^ 24) + CInt(bytes(2) * 2 ^ 16) + CInt(bytes(3) * 2 ^ 8) + CInt(bytes(4) * 2 ^ 0)

        'Console.WriteLine(String.Format("[CameraData] Loading bytes of type {0} and length {1} at time {2}m{3}s{4}", bytes(0), length, Now.Minute.ToString, Now.Second.ToString, Now.Millisecond.ToString))

        Select Case Me._CameraImageType

            Case UvARescue.Agent.CameraImageType.Raw

                SyncLock Me.Mutex

                    Me._Width = CInt(bytes(5) * 2 ^ 8) + CInt(bytes(6))
                    Me._Height = CInt(bytes(7) * 2 ^ 8) + CInt(bytes(8))

                    'copy raw data RGB(1 byte + 1 byte + 1 byte )
                    ReDim Me._RawData(length - 5)
                    Dim raw(length - 5) As Byte 'A clone which can be discarded

                    For i As Integer = 0 To Me._RawData.Length - 1
                        Me._RawData(i) = bytes(i + 9)
                        raw(i) = bytes(i + 9)
                    Next

                    If Not IsNothing(Me._Bitmap) Then
                        Me._Bitmap.Dispose() 'clean up, to prevent memory problems
                        Me._Bitmap = Nothing
                    End If

                    If multiviewMode < 2 AndAlso True Then 'SwitchOff for UsarSkinAgent

                        Dim bmp As Bitmap = Nothing

                        If IsNothing(Me._Bitmap) Then

                            bmp = New Bitmap(Me.Width, Me.Height, PixelFormat.Format24bppRgb)
                        Else
                            bmp = Me._Bitmap
                        End If

                        ' Lock the bitmap's bits.  
                        Dim rect As New Rectangle(0, 0, bmp.Width, bmp.Height)
                        Dim bmpData As System.Drawing.Imaging.BitmapData = bmp.LockBits(rect, _
                            Drawing.Imaging.ImageLockMode.ReadWrite, bmp.PixelFormat)

                        ' Get the address of the first line.
                        Dim ptr As IntPtr = bmpData.Scan0

                        ' Declare an array to hold the bytes of the bitmap.
                        ' This code is specific to a bitmap with 24 bits per pixels.

                        ' Copy the RGB values back to the bitmap
                        System.Runtime.InteropServices.Marshal.Copy(raw, 0, ptr, raw.Length)

                        ' Unlock the bits.
                        bmp.UnlockBits(bmpData)


                        Me._Bitmap = DirectCast(bmp.Clone, Bitmap)
                        bmp.Dispose()

                        'Dim imLength As Integer = CInt(Me.RawData.Length / 3)
                        'Dim red, green, blue As Byte
                        'Dim r(imLength), g(imLength), b(imLength) As Complex
                        'Dim cp As ComplexPart = ComplexPart.Real
                        'Dim size As System.Drawing.Size = New Size(Me.Width, Me.Height) 'If camdata = raw, no bitmap exist

                        'For index As Integer = 0 To imLength - 1

                        '    red = Me.RawData(index * 3)
                        '    green = Me.RawData(index * 3 + 1)
                        '    blue = Me.RawData(index * 3 + 2)

                        '    r(index).Real = red / Byte.MaxValue
                        '    g(index).Real = green / Byte.MaxValue
                        '    b(index).Real = blue / Byte.MaxValue

                        'Next

                        'Me._Bitmap = BitmapConverter.WriteChannel(size, r, cp, g, cp, b, cp) 'MainForm hangs

                    Else
                        'get Multiview subview 
                        'Me._RawData = GetSubView(Me._RawData, viewNumber) Not implemented yet

                    End If


                End SyncLock
            Case UvARescue.Agent.CameraImageType.Png

                SyncLock Me.Mutex

                    If Not IsNothing(Me._Bitmap) Then
                        Me._Bitmap.Dispose() 'clean up, to prevent memory problems
                    End If

                    'Dim png(CInt(1.33 * length)) As Byte
                    'For i As Integer = 0 To bytes.Length - 9 Step 4
                    '    png(i) = bytes(i + 5 + 1)
                    '    png(i + 1) = bytes(i + 5 + 2)
                    '    png(i + 2) = bytes(i + 5)
                    '    png(i + 3) = 0
                    'Next

                    'Dim png(length - 1) As Byte
                    'For i As Integer = 0 To bytes.Length - 8 Step 3
                    '    png(i) = bytes(i + 5)
                    '    png(i + 1) = bytes(i + 5 + 2)
                    '    png(i + 2) = bytes(i + 5 + 1) 'This works OK, only upside down, wrong colours and small region at the right to the left.
                    'Next

                    Dim png(length - 1) As Byte
                    For i As Integer = 0 To bytes.Length - 8 Step 3
                        png(i) = bytes(bytes.Length - 1 - i - 2)  'magneta (0,2,1) 'bluered(0,1,2), 'bluepink(1,0,2), MagnetaPink (2,0,1), GreenPink (1,2,0), 
                        png(i + 1) = bytes(bytes.Length - 1 - i - 1)
                        png(i + 2) = bytes(bytes.Length - 1 - i)
                    Next

                    If Not IsNothing(Me._Bitmap) Then
                        Me._Bitmap.Dispose() 'clean up, to prevent memory problems
                        Me._Bitmap = Nothing
                    End If

                    If multiviewMode < 2 AndAlso True Then 'SwitchOff for UsarSkinAgent

                        Dim bmp As Bitmap = Nothing

                        If IsNothing(Me._Bitmap) Then

                            bmp = New Bitmap(176, 143, PixelFormat.Format24bppRgb) 'Is encoded in png-format, maybe also in bmp
                        Else
                            bmp = Me._Bitmap
                        End If

                        ' Lock the bitmap's bits.  
                        Dim rect As New Rectangle(0, 0, bmp.Width, bmp.Height)
                        Dim bmpData As System.Drawing.Imaging.BitmapData = bmp.LockBits(rect, _
                            Drawing.Imaging.ImageLockMode.ReadWrite, bmp.PixelFormat)

                        ' Get the address of the first line.
                        Dim ptr As IntPtr = bmpData.Scan0

                        ' Declare an array to hold the bytes of the bitmap.
                        ' This code is specific to a bitmap with 24 bits per pixels.

                        ' Copy the RGB values back to the bitmap
                        System.Runtime.InteropServices.Marshal.Copy(png, 0, ptr, png.Length)

                        ' Unlock the bits.
                        bmp.UnlockBits(bmpData)


                        Me._Bitmap = DirectCast(bmp.Clone, Bitmap)
                        bmp.Dispose()
                        Me._Bitmap.RotateFlip(System.Drawing.RotateFlipType.RotateNoneFlipX)

                        Me._Width = Me._Bitmap.Width
                        Me._Height = Me._Bitmap.Height
                    Else
                        'get Multiview subview 
                        'Me._RawData = GetSubView(Me._RawData, viewNumber) Not implemented yet

                    End If

                End SyncLock
            Case Else
                'jpeg
                SyncLock Me.Mutex

                    If Not IsNothing(Me._Bitmap) Then
                        Me._Bitmap.Dispose() 'clean up, to prevent memory problems
                        Me._Bitmap = Nothing
                    End If

                    Dim jpeg(length - 1) As Byte
#If EVEN_SLOWER Then
                    'Code snippet from http://msdn.microsoft.com/en-us/library/ms146631.aspx#Y0
                    ' Initialize unmanged memory to hold the array.
                    Dim size As Integer = Marshal.SizeOf(bytes(0)) * jpeg.Length

                    Dim pnt As IntPtr = Marshal.AllocHGlobal(size)

                    Try
                        ' Copy the array to unmanaged memory.
                        Marshal.Copy(bytes, 5, pnt, jpeg.Length - 1)

                        ' Copy the unmanaged array back to another managed array.

                        Marshal.Copy(pnt, jpeg, 0, jpeg.Length - 1)

                        'Console.WriteLine("The array was copied to unmanaged memory and back.")

                    Finally
                        ' Free the unmanaged memory.
                        Marshal.FreeHGlobal(pnt)
                    End Try
#End If
                    System.Buffer.BlockCopy(bytes, 5, jpeg, 0, jpeg.Length)

                    'original code
                    'For i As Integer = 0 To jpeg.Length - 1
                    '    jpeg(i) = bytes(i + 5)
                    'Next

                    'if Multiview is not in use
                    If multiviewMode < 2 Then
                        Me._Bitmap = DirectCast(New Bitmap(New System.IO.MemoryStream(jpeg)).Clone, Bitmap)

                    Else
                        'get Multiview subview 

                        'Late binding, suggested by http://social.msdn.microsoft.com/Forums/en-US/vbgeneral/thread/e849524d-6c79-47b1-8724-651c8ea7dd93
                        Dim stream As MemoryStream = New System.IO.MemoryStream(jpeg)
                        Dim subView As Bitmap = New Bitmap(stream, False)


                        'Dim subView As Bitmap
                        'subView = DirectCast(New Bitmap(New System.IO.MemoryStream(jpeg)).Clone, Bitmap)

                        Me._Bitmap = GetSubView(subView, viewNumber, multiviewMode)

                        subView.Dispose()

                    End If

                    Me._Width = Me._Bitmap.Width
                    Me._Height = Me._Bitmap.Height

                    jpeg = Nothing

                End SyncLock

        End Select
        bytes = Nothing

    End Sub

    'subview getter function for Raw data
    Private Function GetSubView(ByVal rawData As Byte(), ByVal spawnNumber As Integer) As Byte()

        'TODO: Raw subview getter function should be implemented here

        Dim subViewRaw As Byte() = Nothing
        Return subViewRaw
    End Function


    'subview getter function for Jpeg data
    Private Function GetSubView(ByVal bMap As Bitmap, ByVal spawnNumber As Integer, ByVal multiviewMode As Byte) As Bitmap

        Try
            'hardcopy, suggested by http://bytes.com/topic/visual-basic-net/answers/371925-new-bitmap-very-slow
            'Dim fullView As Bitmap = New Bitmap(bMap.Width, bMap.Height)
            'Dim grx As Graphics = Graphics.FromImage(bMap)
            'grx.DrawImage(fullView, 0, 0)
            'grx.Dispose()

            Dim fullView As Bitmap = New Bitmap(bMap)

            'saving cameradata to file
            'Dim pathName, imageFile As String
            'pathName = Path.Combine(My.Application.Info.DirectoryPath, "Images")
            'imageFile = Path.Combine(pathName, String.Concat(CStr(spawnNumber), "_FULL.jpg"))
            'Dim fileInfo As New FileInfo(imageFile)
            'If Not fileInfo.Directory.Exists Then
            '   FileInfo.Directory.Create()
            'End If
            'fullView.Save(imageFile, System.Drawing.Imaging.ImageFormat.Jpeg)

            Dim row, column, x, y, width, height As Integer

            'row = (spawnNumber - 1) Mod multiviewMode
            'column = (spawnNumber - 1) \ multiviewMode

            row = (spawnNumber) Mod multiviewMode
            column = (spawnNumber) \ multiviewMode

            If row < 0 Then
                Console.WriteLine("[GetSubView]: Warning row < 0, set to 0")
                row = 0
            End If
            If column < 0 Then
                Console.WriteLine("[GetSubView]: Warning column < 0, set to 0")
                column = 0
            End If

            width = fullView.Width \ multiviewMode
            height = fullView.Height \ multiviewMode

            'Console.WriteLine("full: {0},{1}, cropped: {2},{3}", fullView.Width, fullView.Height, width, height)

            x = row * width
            y = column * height

            Dim cloneRect As New Drawing.Rectangle(x, y, width, height)
            Dim format As Imaging.PixelFormat = fullView.PixelFormat
            Dim subView As Bitmap = fullView.Clone(cloneRect, format)

            'saving cameradata to file
            'imageFile = Path.Combine(pathName, String.Concat(CStr(spawnNumber), "_SUBVIEW.jpg"))
            'subView.Save(imageFile, System.Drawing.Imaging.ImageFormat.Jpeg)

            Return subView
        Catch ex As Exception
            Console.WriteLine("Error occurred while trying to create a subview.")
            Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace)
            Return bMap

        End Try

    End Function

    Public Sub Load(ByVal msg As System.Collections.Specialized.StringDictionary) Implements ISensorData.Load
    End Sub

    Private SlowConvertMethod As Boolean = False

    Private Sub ConvertFromBitmap()

        'copy data from Me._Bitmap to Me._RawData

        ReDim Me._RawData(Me._Width * Me._Height * 3)

        If SlowConvertMethod = True AndAlso Me._Bitmap.PixelFormat = PixelFormat.Format24bppRgb Then

            Dim bits As BitmapData = Me._Bitmap.LockBits(New Rectangle(0, 0, Me._Bitmap.Width, Me._Bitmap.Height), ImageLockMode.ReadOnly, Me._Bitmap.PixelFormat)

            'From example http://www.bobpowell.net/mapfill.aspx

            Dim offset As Integer = 0
            Dim index As Integer = 0
            For x As Integer = 0 To Me.Width - 1
                For y As Integer = 0 To Me.Height - 1

                    offset = y * bits.Stride + (3 * x) '3 colors x 8 bits = 24bbp
                    _RawData(index) = Marshal.ReadByte(bits.Scan0, offset) 'red
                    _RawData(index + 1) = Marshal.ReadByte(bits.Scan0, offset + 1) 'green
                    _RawData(index + 2) = Marshal.ReadByte(bits.Scan0, offset + 2) 'blue

                    index += 3
                Next 'y
            Next 'x

            Me._Bitmap.UnlockBits(bits)

        ElseIf SlowConvertMethod = True AndAlso Me._Bitmap.PixelFormat = PixelFormat.Format32bppArgb Then

            Dim bits As BitmapData = Me._Bitmap.LockBits(New Rectangle(0, 0, Me._Bitmap.Width, Me._Bitmap.Height), ImageLockMode.ReadOnly, Me._Bitmap.PixelFormat)

            'From example http://www.bobpowell.net/mapfill.aspx

            Dim offset As Integer = 0
            Dim index As Integer = 0
            For x As Integer = 0 To Me.Width - 1
                For y As Integer = 0 To Me.Height - 1

                    offset = y * bits.Stride + (4 * x) 'skip alpha

                    _RawData(index) = Marshal.ReadByte(bits.Scan0, offset) 'red
                    _RawData(index + 1) = Marshal.ReadByte(bits.Scan0, offset + 1) 'green
                    _RawData(index + 2) = Marshal.ReadByte(bits.Scan0, offset + 2) 'blue

                    index += 3
                Next 'y
            Next 'x

            Me._Bitmap.UnlockBits(bits)

        Else
            'Fast, but memory problems for larger resolutions

            Dim R(Me._Width * Me._Height) As Complex
            Dim G(Me._Width * Me._Height) As Complex
            Dim B(Me._Width * Me._Height) As Complex

            Try
                MathNet.SignalProcessing.DataSources.BitmapConverter.ReadChannel(Me._Bitmap, R, G, B)
            Catch ex As Exception
                Console.WriteLine("Error occurred while trying to read RGB channels.")
                Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace)
            End Try

            'ReDim Me._RawData(R.Length * 3) ' already done above

            Dim index As Integer = 0

            Try
                For i As Integer = 0 To (R.Length) - 1
                    _RawData(index) = CByte(R(i).Real * 256)
                    _RawData(index + 1) = CByte(G(i).Real * 256)
                    _RawData(index + 2) = CByte(B(i).Real * 256)
                    index += 3
                Next
            Catch ex As Exception
                Console.WriteLine("Error converting RGB data to RAW image")
                Console.WriteLine(ex.Message & vbNewLine & ex.StackTrace)
            End Try


        End If
    End Sub

#End Region

End Class
