Imports System.Math

<Serializable()> _
Public Class Pose3D

#Region " Constructor and Properties "

    Private _Position As Vector3
    Private _Rotation As Vector3

    Public Sub New()
        Me.New(New Vector3(0, 0, 0), New Vector3(0, 0, 0))
    End Sub

    Public Sub New(ByVal position As Vector3, ByVal rotation As Vector3)
        Me.Position = position
        Me.Rotation = rotation
    End Sub

    Public Sub New(ByVal startLocation As String, ByVal startRotation As String)
        Dim parts() As String = Strings.Split(startLocation, ",")

        If parts.Length = 3 Then
            Me.Position = New Vector3(Single.Parse(parts(0)), Single.Parse(parts(1)), Single.Parse(parts(2)))
        Else
            Me.Position = New Vector3(0, 0, 0)
        End If


        parts = Strings.Split(startRotation, ",")
        If parts.Length = 3 Then
            Me.Rotation = New Vector3(Single.Parse(parts(0)), Single.Parse(parts(1)), Single.Parse(parts(2)))
        Else
            Me.Rotation = New Vector3(0, 0, 0)
        End If

    End Sub


    Public Sub New(ByVal X As Double, ByVal Y As Double, ByVal Z As Double, ByVal Roll As Double, ByVal Pitch As Double, ByVal Yaw As Double)

        Me.New()

        Me.Position.X = X
        Me.Position.Y = Y
        Me.Position.Z = Z

        Me.Rotation.X = Roll
        Me.Rotation.Y = Pitch
        Me.Rotation.Z = Yaw

    End Sub

    Public Property Position() As Vector3
        Get
            Return _Position
        End Get
        Set(ByVal value As Vector3)
            Me._Position = value
        End Set
    End Property

    Public Property X() As Double
        Get
            Return Me._Position.X
        End Get
        Set(ByVal value As Double)
            Me._Position.X = value
        End Set
    End Property

    Public Property Y() As Double
        Get
            Return Me._Position.Y
        End Get
        Set(ByVal value As Double)
            Me._Position.Y = value
        End Set
    End Property

    Public Property Z() As Double
        Get
            Return Me._Position.Z
        End Get
        Set(ByVal value As Double)
            Me._Position.Z = value
        End Set
    End Property

    Public Property Rotation() As Vector3
        Get
            Return _Rotation
        End Get
        Set(ByVal value As Vector3)
            Me._Rotation = value
        End Set
    End Property

    Public Property Roll() As Double
        Get
            Return Me._Rotation.X
        End Get
        Set(ByVal value As Double)
            Me._Rotation.X = value
        End Set
    End Property

    Public Property Pitch() As Double
        Get
            Return Me._Rotation.Y
        End Get
        Set(ByVal value As Double)
            Me._Rotation.Y = value
        End Set
    End Property

    Private _Yaw As Double
    Public Property Yaw() As Double
        Get
            Return Me._Rotation.Z
        End Get
        Set(ByVal value As Double)
            Me._Rotation.Z = value
        End Set
    End Property

    Public Property Pose As Pose2D
        Get
            Return New Pose2D(Me._Position.X, Me._Position.Y, Me._Rotation.Z)
        End Get
        Set(ByVal value As Pose2D)
            If Not IsNothing(value) Then
                Me._Position.X = value.X
                Me._Position.Y = value.Y
                Me._Rotation.Z = value.Rotation
            End If

        End Set

    End Property


#End Region

#Region " Smart Functions "


    Public Function ToGlobalMatrix() As TMatrix3D
        Return New TMatrix3D(Me.Position, Me.Rotation)
    End Function
#If NEEDED Then
    Public Function ToLocalMatrix() As TMatrix2D
        Dim mx As New TMatrix2D(-Me.Rotation)
        Dim merotated As Pose2D = mx * Me
        mx.Translation = -merotated.Position
        Return mx
    End Function
#End If
    Public Function ToGlobal(ByVal currentOrigin As Pose3D) As Pose3D
        Dim rotmx As TMatrix3D = currentOrigin.ToGlobalMatrix
        Dim pglobal As Pose3D = rotmx * Me
        Return pglobal
    End Function
#If Needed Then
    Public Function ToLocal(ByVal targetOrigin As Pose2D) As Pose2D
        Dim rotmx As TMatrix2D = targetOrigin.ToLocalMatrix
        Dim plocal As Pose2D = rotmx * Me
        Return plocal
    End Function
#End If

#End Region

#Region " Normalize Rotation "

    Public Function GetNormalizedRotation() As Double
        Dim radians As Double = Me.Yaw
        While radians > PI
            radians -= 2 * PI
        End While
        While radians <= -PI
            radians += 2 * PI
        End While
        Return radians
    End Function

#End Region
End Class

