Imports System.Math

<Serializable()> _
Public Class TMatrix3D
    Inherits Matrix4

#Region " Constructors "

    Public Sub New()
        Me.New(0, 0, 0, 0, 0, 0)
    End Sub
#If Needed Then
    Public Sub New(ByVal rotation As Vector3)
        Me.New(0, 0, 0, rotation.X, rotation.Y, rotation.Z)
    End Sub

    Public Sub New(ByVal translation As Vector3)
        Me.New(translation.X, translation.Y, translation.Z, 0, 0, 0)
    End Sub
#End If

    Public Sub New(ByVal translation As Vector3, ByVal rotation As Vector3)
        Me.New(translation.X, translation.Y, translation.Z, rotation.X, rotation.Y, rotation.Z)
    End Sub
#If NEEDED Then
    Public Sub New(ByVal dx As Double, ByVal dy As Double)
        Me.New(dx, dy, 0)
    End Sub
#End If

    Public Sub New(ByVal dx As Double, ByVal dy As Double, ByVal dz As Double, ByVal roll As Double, ByVal pitch As Double, ByVal yaw As Double)
        Me.LoadIdentity()
        Me.DeltaX = dx
        Me.DeltaY = dy
        Me.DeltaZ = dz
        Me.Roll = roll
        Me.pitch = pitch
        Me.Yaw = yaw
    End Sub

#End Region


#Region " Data Wrappers "

    Public Property DeltaX() As Double
        Get
            Return Me(0, 3)
        End Get
        Set(ByVal value As Double)
            Me(0, 3) = value
        End Set
    End Property
    Public Property DeltaY() As Double
        Get
            Return Me(1, 3)
        End Get
        Set(ByVal value As Double)
            Me(1, 3) = value
        End Set
    End Property
    Public Property DeltaZ() As Double
        Get
            Return Me(2, 3)
        End Get
        Set(ByVal value As Double)
            Me(2, 3) = value
        End Set
    End Property

    Private _Roll As Double
    Private SinX As Double 'no underscore, confusing with minus
    Private CosX As Double
    Public Property Roll() As Double
        Get
            Return _Roll
        End Get
        Set(ByVal value As Double)
            If Not value = Me._Roll Then
                Me._Roll = value
                Me.SinX = CType(Sin(value), Double)
                Me.CosX = CType(Cos(value), Double)

                Me(0, 0) = CosY * CosZ 'from http://en.wikipedia.org/wiki/3D_projection#Perspective_projection
                Me(0, 1) = CosY * SinZ
                Me(0, 2) = -SinY
                Me(1, 0) = SinY * SinX * CosZ - CosX * SinZ
                Me(1, 1) = SinX * SinY * SinZ + CosX * CosX
                Me(1, 2) = SinX * CosY
                Me(2, 0) = CosX * SinY * CosZ + SinX * SinZ
                Me(2, 1) = CosX * SinY * SinZ - SinX * CosZ
                Me(2, 2) = CosX * CosY

                ' checked against http://sourceforge.net/mailarchive/forum.php?thread_name=fb6820a0808291957u693724e3j900243e6b88ed451%40mail.gmail.com&forum_name=glsdk-devel
                ' to be tested (could be mirrored)

            End If
        End Set
    End Property

    Private _Pitch As Double
    Private SinY As Double 'no underscore, confusing with minus
    Private CosY As Double
    Public Property Pitch() As Double
        Get
            Return _Pitch
        End Get
        Set(ByVal value As Double)
            If Not value = Me._Pitch Then
                Me._Pitch = value
                Me.SinY = CType(Sin(value), Double)
                Me.CosY = CType(Cos(value), Double)

                Me(0, 0) = CosY * CosZ 'from http://en.wikipedia.org/wiki/3D_projection#Perspective_projection
                Me(0, 1) = CosY * SinZ
                Me(0, 2) = -SinY
                Me(1, 0) = SinY * SinX * CosZ - CosX * SinZ
                Me(1, 1) = SinX * SinY * SinZ + CosX * CosX
                Me(1, 2) = SinX * CosY
                Me(2, 0) = CosX * SinY * CosZ + SinX * SinZ
                Me(2, 1) = CosX * SinY * SinZ - SinX * CosZ
                Me(2, 2) = CosX * CosY

            End If
        End Set
    End Property

    Private _Yaw As Double
    Private SinZ As Double 'no underscore, confusing with minus
    Private CosZ As Double
    Public Property Yaw() As Double
        Get
            Return _Yaw
        End Get
        Set(ByVal value As Double)
            If Not value = Me._Yaw Then
                Me._Yaw = value
                Me.SinZ = CType(Sin(value), Double)
                Me.CosZ = CType(Cos(value), Double)

                Me(0, 0) = CosY * CosZ 'from http://en.wikipedia.org/wiki/3D_projection#Perspective_projection
                Me(0, 1) = CosY * SinZ
                Me(0, 2) = -SinY
                Me(1, 0) = SinY * SinX * CosZ - CosX * SinZ
                Me(1, 1) = SinX * SinY * SinZ + CosX * CosX
                Me(1, 2) = SinX * CosY
                Me(2, 0) = CosX * SinY * CosZ + SinX * SinZ
                Me(2, 1) = CosX * SinY * SinZ - SinX * CosZ
                Me(2, 2) = CosX * CosY
            End If
        End Set
    End Property
#If NEEDED Then
    Public Property Translation() As Vector2
        Get
            Return New Vector2(Me.DeltaX, Me.DeltaY)
        End Get
        Set(ByVal value As Vector2)
            Me.DeltaX = value.X
            Me.DeltaY = value.Y
        End Set
    End Property
#End If

#End Region

#If NEEDED Then
#Region " Special Operators "

    Public Overloads Sub Translate(ByVal translation As Vector2)
        Me.Translate(translation.X, translation.Y)
    End Sub
    Public Overloads Sub Translate(ByVal x As Double, ByVal y As Double)
        Me.DeltaX += x
        Me.DeltaY += y
    End Sub

    Public Overloads Sub Rotate(ByVal angle As Double)
        Me.Rotation += angle
    End Sub

#End Region
#End If

#Region " Operators "

    Public Overloads Function Multiply(ByVal vector As Vector2) As Vector2
        Dim result As New Vector2()
        result.X = vector.X * Me(0, 0) + vector.Y * Me(0, 1) + 0 * Me(0, 2) + Me(0, 3)
        result.Y = vector.X * Me(1, 0) + vector.Y * Me(1, 1) + 0 * Me(1, 2) + Me(1, 3)
        Return result
    End Function

    Public Overloads Function Multiply(ByVal vector As Vector3) As Vector3
        Dim result As New Vector3()
        result.X = vector.X * Me(0, 0) + vector.Y * Me(0, 1) + vector.Z * Me(0, 2) + Me(0, 3)
        result.Y = vector.X * Me(1, 0) + vector.Y * Me(1, 1) + vector.Z * Me(1, 2) + Me(1, 3)
        result.Z = vector.X * Me(2, 0) + vector.Y * Me(2, 1) + vector.Z * Me(2, 2) + Me(2, 3)
        Return result
    End Function

    Public Overloads Function Multiply(ByVal vector As Vector4) As Vector4
        Dim result As New Vector4()
        result.X = vector.X * Me(0, 0) + vector.Y * Me(0, 1) + vector.Z * Me(0, 2) + vector.Scale * Me(0, 3)
        result.Y = vector.X * Me(1, 0) + vector.Y * Me(1, 1) + vector.Z * Me(1, 2) + vector.Scale * Me(1, 3)
        result.Z = vector.X * Me(2, 0) + vector.Y * Me(2, 1) + vector.Z * Me(2, 2) + vector.Scale * Me(2, 3)
        result.Scale = vector.X * Me(3, 0) + vector.Y * Me(3, 1) + vector.Z * Me(3, 2) + vector.Scale * Me(3, 3)
        Return result
    End Function

    Public Overloads Shared Operator -(ByVal right As TMatrix3D) As TMatrix3D
        Return New TMatrix3D(-right.DeltaX, -right.DeltaY, -right.DeltaZ, -right.Roll, -right.Pitch, -right.Yaw)
    End Operator

    Public Overloads Shared Operator *(ByVal matrix As TMatrix3D, ByVal vector As Vector4) As Vector4
        Return matrix.Multiply(vector)
    End Operator

    Public Overloads Shared Operator *(ByVal matrix As TMatrix3D, ByVal vector As Vector3) As Vector3
        Return matrix.Multiply(vector)
    End Operator
    Public Overloads Shared Operator *(ByVal matrix As TMatrix3D, ByVal vector As Vector2) As Vector2
        Return New Vector2(matrix.Multiply(vector))
    End Operator
    Public Overloads Shared Operator *(ByVal matrix As TMatrix3D, ByVal pose As Pose3D) As Pose3D
        Dim vector As New Vector4(pose.X, pose.Y, pose.Z, 1)
        vector = matrix * vector
        Return New Pose3D(vector.X, vector.Y, vector.Z, pose.Roll + matrix.Roll, pose.Pitch + matrix.Pitch, pose.Yaw + matrix.Yaw)
    End Operator

#End Region


End Class

